from entmoot.models.enting import Enting
from entmoot.models.model_params import (
    EntingParams,
    TrainParams,
    TreeTrainParams,
    UncParams,
)
from entmoot.optimizers.gurobi_opt import GurobiOptimizer
from entmoot.optimizers.pyomo_opt import PyomoOptimizer
from entmoot.problem_config import ProblemConfig as ProblemConfig



from abc import ABC, abstractmethod
from typing import Generic, List, Literal, Optional, TypeVar

import gurobipy
import numpy as np
import pyomo.environ as pyo

from entmoot.typing.optimizer_stubs import GurobiModelT, PyomoModelT

B = TypeVar("B", int, float)
BoundsT = tuple[float, float]
CategoriesT = list[str | float | int] | tuple[str | float | int, ...]


class FeatureType(ABC):
    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def get_enc_bnds(self):
        pass

    def is_real(self):
        return False

    def is_cat(self):
        return False

    def is_int(self):
        return False

    def is_bin(self):
        return False

    def encode(self, xi):
        return xi

    def decode(self, xi):
        return xi

    @abstractmethod
    def sample(self, num_samples: int, rng: np.random.Generator):
        pass

    @abstractmethod
    def get_pyomo_domain_and_bounds(self, i: int) -> tuple[dict, dict]:
        pass

    @abstractmethod
    def get_gurobi_variable(self, model: GurobiModelT):
        pass

    def __str__(self):
        return f"{self.name} :: {self.__class__.__name__}"


class Ordinal(FeatureType, Generic[B]):
    pyomo_domain: pyo.Set
    gurobi_vtype: str

    def __init__(self, lb: B, ub: B, name: str):
        super().__init__(name)
        self.lb = lb
        self.ub = ub

    def get_enc_bnds(self):
        return (self.lb, self.ub)

    def get_pyomo_domain_and_bounds(self, i: int) -> tuple[dict, dict]:
        return {i: self.pyomo_domain}, {i: self.get_enc_bnds()}

    def get_gurobi_variable(self, model: GurobiModelT):
        return model.addVar(
            lb=self.lb, ub=self.ub, name=self.name, vtype=self.gurobi_vtype
        )

    def __str__(self):
        return f"{super().__str__()} :: {self.get_enc_bnds()}"


class Real(Ordinal[float]):
    pyomo_domain = pyo.Reals
    gurobi_vtype = "C"

    def is_real(self):
        return True

    def sample(self, num_samples, rng):
        return rng.uniform(low=self.lb, high=self.ub, size=num_samples)


class Integer(Ordinal[int]):
    pyomo_domain = pyo.Integers
    gurobi_vtype = "I"

    def is_int(self):
        return True

    def decode(self, xi):
        return int(xi)

    def sample(self, num_samples, rng):
        return rng.integers(low=self.lb, high=self.ub, size=num_samples, endpoint=True)


class Binary(Ordinal[int]):
    pyomo_domain = pyo.Binary
    gurobi_vtype = "B"

    def __init__(self, name: str):
        super().__init__(0, 1, name)

    def decode(self, xi):
        return abs(int(xi))

    def is_bin(self):
        return True

    def sample(self, num_samples, rng):
        return rng.integers(low=self.lb, high=self.ub, size=num_samples, endpoint=True)


class Categorical(FeatureType):
    def __init__(self, cat_list: CategoriesT, name: str):
        super().__init__(name)
        self._cat_list = cat_list

        # encode categories
        self._enc2str, self._str2enc = {}, {}
        self._enc_cat_list = []
        for enc, cat in enumerate(cat_list):
            self._enc_cat_list.append(enc)
            self._enc2str[enc] = cat
            self._str2enc[cat] = enc

    def get_enc_bnds(self):
        return self._enc_cat_list

    @property
    def cat_list(self):
        return self._cat_list

    @property
    def enc_cat_list(self):
        return self._enc_cat_list

    def encode(self, xi):
        return self._str2enc[xi]

    def decode(self, xi):
        return self._enc2str[xi]

    def is_cat(self):
        return True

    def sample(self, num_samples, rng):
        return rng.integers(0, len(self.cat_list), size=num_samples)

    def get_pyomo_domain_and_bounds(self, i: int):
        # We encode the index of this variable by (i, enc, cat), where 'i' is the position in the list of
        # features, 'enc' is the corresponding encoded numerical value and 'cat' is the category
        return {
            (i, enc, cat): pyo.Binary
            for (enc, cat) in zip(self.enc_cat_list, self.cat_list)
        }, {}

    def get_gurobi_variable(self, model: GurobiModelT):
        return {
            enc: model.addVar(name=f"{self.name}_{cat}", vtype="B")
            for (enc, cat) in zip(self.enc_cat_list, self.cat_list)
        }


class Objective:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"{self.name} :: {self.__class__.__name__}"


class MinObjective(Objective):
    sign = 1


class MaxObjective(Objective):
    sign = -1


AnyFeatureT = Real | Integer | Categorical | Binary
FeatureT = TypeVar("FeatureT", bound=FeatureType)
AnyObjectiveT = MinObjective | MaxObjective


class ProblemConfig:
    def __init__(self, rnd_seed: Optional[int] = None):
        self._feat_list = []
        self._obj_list = []
        self._rnd_seed = rnd_seed
        self.rng = np.random.default_rng(rnd_seed)

    @property
    def cat_idx(self):
        return tuple([i for i, feat in self.get_idx_and_feat_by_type(Categorical)])

    @property
    def non_cat_idx(self):
        return tuple([i for i, feat in self.get_idx_and_feat_by_type(Ordinal)])

    @property
    def non_cat_lb(self):
        return tuple([feat.lb for i, feat in self.get_idx_and_feat_by_type(Ordinal)])

    @property
    def non_cat_ub(self):
        return tuple([feat.ub for i, feat in self.get_idx_and_feat_by_type(Ordinal)])

    @property
    def non_cat_bnd_diff(self):
        return tuple(
            [feat.ub - feat.lb for i, feat in self.get_idx_and_feat_by_type(Ordinal)]
        )

    def get_idx_and_feat_by_type(
        self, feature_type: type[FeatureT]
    ) -> tuple[tuple[int, FeatureT], ...]:
        return tuple(
            [
                (i, feat)
                for i, feat in enumerate(self.feat_list)
                if isinstance(feat, feature_type)
            ]
        )

    @property
    def feat_list(self) -> list[AnyFeatureT]:
        return self._feat_list

    @property
    def obj_list(self) -> list[AnyObjectiveT]:
        return self._obj_list

    @property
    def rnd_seed(self):
        return self._rnd_seed

    def get_enc_bnd(self):
        return [feat.get_enc_bnds() for feat in self.feat_list]

    def _encode_xi(self, xi: List):
        return np.asarray(
            [feat.encode(xi[idx]) for idx, feat in enumerate(self.feat_list)]
        )

    def _decode_xi(self, xi: List):
        return [feat.decode(xi[idx]) for idx, feat in enumerate(self.feat_list)]

    def encode(self, X: list) -> np.ndarray:
        if len(X) == 0:
            return np.array([])
        elif len(X) == 1:
            return self._encode_xi(X[0])
        else:
            enc = [self._encode_xi(xi) for xi in X]
            return np.asarray(enc)

    def decode(self, X: list | np.ndarray) -> list[float | int | str] | np.ndarray:
        if len(X) == 0:
            return []
        elif len(X) == 1:
            return self._decode_xi(X[0])
        else:
            dec = [self._decode_xi(xi) for xi in X]
            # TODO: change to list for consistency
            return np.asarray(dec)

    def add_feature(
        self,
        feat_type: Literal["real", "integer", "binary", "categorical"],
        bounds: Optional[BoundsT | CategoriesT] = None,
        name: Optional[str] = None,
    ):
        # TODO: Move validation logic to Features
        if name is None:
            name = f"feat_{len(self.feat_list)}"

        if feat_type == "binary":
            self._feat_list.append(Binary(name=name))
            return

        if bounds is None:
            raise ValueError(
                "Please provide bounds for feature types in '(real, integer, categorical)'"
            )

        # perform basic input checks and define features of space
        if feat_type in ("real", "integer"):
            assert len(bounds) == 2, (
                f"Too many values for bounds for feat_type '{feat_type}'. Expected '2', "
                f"got '{len(bounds)}'. Check feature '{name}'"
            )

            lb, ub = bounds

            if feat_type == "real":
                if isinstance(lb, int):
                    lb = float(lb)
                if isinstance(ub, int):
                    ub = float(ub)

                assert isinstance(lb, float) and isinstance(ub, float), (
                    f"Wrong type for bounds of feat_type '{feat_type}'. Expected '(float, "
                    f"float)', got '({type(lb)}, {type(ub)})'. Check feature '{name}'"
                )

                assert lb < ub, (
                    f"Lower bound '{lb}' of feat_type '{feat_type}' needs to be "
                    f"smaller than upper bound. Check feature '{name}'."
                )

                self._feat_list.append(Real(lb=lb, ub=ub, name=name))
            else:
                assert isinstance(lb, int) and isinstance(ub, int), (
                    f"Wrong type for bounds of feat_type '{feat_type}'. Expected '(int, int)', "
                    f"got '({type(lb)}, {type(ub)})'. Check feature '{name}'"
                )

                assert lb < ub, (
                    f"Lower bound '{lb}' of feat_type '{feat_type}' needs to be "
                    f"smaller than upper bound. Check feature '{name}'."
                )

                self._feat_list.append(Integer(lb=lb, ub=ub, name=name))

        elif feat_type == "categorical":
            assert len(bounds) > 1, (
                f"Not enough categories specified for bounds of feat_type '{feat_type}'. "
                f"Expected 'len(bounds) > 1', got 'len(bounds) = {len(bounds)}'. "
                f"Check feature '{name}'."
            )

            assert all(isinstance(item, str) for item in bounds), (
                f"Wrong type for bounds of feat_type '{feat_type}'. Expected 'str' for all "
                f"categories."
            )

            assert len(bounds) == len(
                set(bounds)
            ), f"Categories of feat_type '{feat_type}' are not all unique."

            self._feat_list.append(Categorical(cat_list=bounds, name=name))  # type: ignore

        else:
            raise ValueError(
                f"No support for feat_type '{feat_type}'. Check feature '{name}'."
            )

    def add_min_objective(self, name: Optional[str] = None):
        if name is None:
            name = f"obj_{len(self.obj_list)}"

        self._obj_list.append(MinObjective(name=name))

    def add_max_objective(self, name: Optional[str] = None):
        if name is None:
            name = f"obj_{len(self.obj_list)}"

        self._obj_list.append(MaxObjective(name=name))

    def transform_objective(self, y: np.ndarray) -> np.ndarray:
        """Transform data for minimisation/maximisation"""
        # y.shape = (num_samples, num_obj)
        signs = np.array([obj.sign for obj in self.obj_list]).reshape(1, -1)
        return y * signs

    def get_rnd_sample_numpy(self, num_samples):
        # returns np.array for faster processing
        array_list = [feat.sample(num_samples, self.rng) for feat in self.feat_list]
        return np.squeeze(np.column_stack(array_list))

    def get_rnd_sample_list(self, num_samples=1, cat_enc=False):
        # returns list of tuples
        sample_list = self.get_rnd_sample_numpy(num_samples).tolist()
        if not cat_enc:
            for n, sample in enumerate(sample_list):
                sample_list[n] = [
                    feat.decode(sample[i]) for i, feat in enumerate(self.feat_list)
                ]
        return sample_list if len(sample_list) > 1 else sample_list[0]

    def get_gurobi_model_core(self, env=None) -> GurobiModelT:
        import gurobipy as gur

        # initialize gurobi model
        model: GurobiModelT = gur.Model(env=env)  # type: ignore

        # initalize var space
        model._all_feat = [feat.get_gurobi_variable(model) for feat in self.feat_list]

        model.update()
        return model

    def copy_gurobi_model_core(self, model_core: GurobiModelT) -> GurobiModelT:
        """
        Computes a copy of a Gurobi model which is decoupled from the original, i.e. you can modify the copy without
        changing the original model.
        :param model_core: The gurobi model we want to copy
        :type model_core: gurobi.Model
        :return:
        :rtype: gurobi.Model
        """
        copy_model_core: GurobiModelT = model_core.copy()  # type: ignore
        # Since Gurobi's .copy() method does not copy attributes like ._all_feat we have to copy them manually
        copy_model_core._all_feat = []

        # transfer feature var list to model copy
        copy_all_feat = []
        for var in model_core._all_feat:
            if isinstance(var, gurobipy.Var):
                # ordinal variable
                copy_all_feat.append(copy_model_core.getVarByName(var.VarName))

            else:
                # categorical variable
                copy_all_feat.append(
                    {
                        enc: copy_model_core.getVarByName(v.VarName)
                        for (enc, v) in var.items()
                    }
                )

        copy_model_core._all_feat = copy_all_feat

        return copy_model_core

    def get_pyomo_model_core(self) -> PyomoModelT:
        import pyomo.environ as pyo

        # initialize Pyomo model
        model: PyomoModelT = pyo.ConcreteModel()

        # Initialize var space
        model._all_feat = []

        # Define some auxiliary dictionaries
        index_to_var_domain = {}
        index_to_var_bounds = {}
        for i, feat in enumerate(self.feat_list):
            domain, bounds = feat.get_pyomo_domain_and_bounds(i)
            index_to_var_domain.update(domain)
            index_to_var_bounds.update(bounds)

        # Build Pyomo index set from dictionary keys
        # Why strings?? Well, pyomo doesn't support lists where some indices are multidimensional, but this is handy
        # for encoded categorical features
        model.indices_features = pyo.Set(
            initialize=[str(x) for x in index_to_var_domain]
        )

        # These auxiliary functions are needed, because Pyomo does not accept the dictionaries instead. The 'model'
        # argument is needed as well, but not explicitly used.
        # We hve the eval-statement eval(i) instead of just i, because the indices are strings for reasons explained
        # above :(
        def i_to_dom(model, i):
            return index_to_var_domain[eval(i)]

        def i_to_bounds(model, i):
            return index_to_var_bounds.get(eval(i), (None, None))

        # Define decision variables for Pyomo model
        model.x = pyo.Var(model.indices_features, domain=i_to_dom, bounds=i_to_bounds)
        # We store the decision variables corresponding to the features in the list _all_feat in order to follow a
        # similar strategy to the gurobi model. Can be replaced in the future by a more direct approach.
        for i_str in model.x:
            i = eval(i_str)
            if isinstance(i, int):
                model._all_feat.append(model.x[i_str])
            elif isinstance(i, tuple):
                if i[1] == 0:
                    model._all_feat.append({})

                model._all_feat[-1][i[1]] = model.x[i_str]
            else:
                raise TypeError

        return model

    def copy_pyomo_model_core(self, model_core: PyomoModelT) -> PyomoModelT:
        """
        Computes a copy of a Pyomo model which is decoupled from the original, i.e. you can modify the copy without
        changing the original model.
        :param model_core: A Pyomo model
        :type model_core: pyomo.environ.ConcreteModel
        :return: A copy of the Pyomo model
        :rtype: pyomo.environ.ConcreteModel
        """
        return model_core.clone()

    def __str__(self):
        return "\n".join(
            (
                "\nPROBLEM SUMMARY",
                "-" * 15,
                "features:",
                *map(str, self.feat_list),
                "\nobjectives:",
                *map(str, self.obj_list),
            )
        )



import pyomo.environ as pyo
import pytest

from entmoot.benchmarks import (
    build_multi_obj_categorical_problem,
    build_reals_only_problem,
    eval_multi_obj_cat_testfunc,
    eval_reals_only_testfunc,
)
from entmoot.constraints import (
    ConstraintList,
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
)
from entmoot.models.enting import Enting
from entmoot.models.model_params import EntingParams, UncParams
from entmoot.optimizers.pyomo_opt import PyomoOptimizer
from entmoot.problem_config import ProblemConfig

PARAMS = EntingParams(unc_params=UncParams(dist_metric="l1", acq_sense="exploration"))


def test_linear_equality_constraint():
    problem_config = ProblemConfig(rnd_seed=73)
    # number of objectives
    number_objectives = 2
    build_multi_obj_categorical_problem(problem_config, n_obj=number_objectives)

    # sample data
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=number_objectives)

    enting = Enting(problem_config, params=PARAMS)
    # fit tree ensemble
    enting.fit(rnd_sample, testfunc_evals)

    model_pyo = problem_config.get_pyomo_model_core()
    # define the constraint
    # then immediately apply it to the model
    model_pyo.xy_equal = LinearEqualityConstraint(
        feature_keys=["feat_3", "feat_4"], coefficients=[1, -1], rhs=0
    ).as_pyomo_constraint(model_pyo, problem_config.feat_list)

    model_pyo.yz_equal = LinearEqualityConstraint(
        feature_keys=["feat_4", "feat_5"], coefficients=[1, -1], rhs=0
    ).as_pyomo_constraint(model_pyo, problem_config.feat_list)

    # optimise the model
    params_pyomo = {"solver_name": "gurobi"}
    opt_pyo = PyomoOptimizer(problem_config, params=params_pyomo)
    res_pyo = opt_pyo.solve(enting, model_core=model_pyo)
    x_opt, y_opt, z_opt = res_pyo.opt_point[3:]

    assert round(x_opt, 5) == round(y_opt, 5) and round(y_opt, 5) == round(z_opt, 5)


@pytest.mark.parametrize(
    "min_count,max_count",
    [
        (1, 3),
        (0, 5),
        (1, 1),
        (5, 5),
    ],
)
def test_nchoosek_constraint(min_count, max_count):
    # standard setting up of problem
    problem_config = ProblemConfig(rnd_seed=73)
    build_reals_only_problem(problem_config)
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=50)
    testfunc_evals = eval_reals_only_testfunc(rnd_sample)

    enting = Enting(problem_config, params=PARAMS)
    # fit tree ensemble
    enting.fit(rnd_sample, testfunc_evals)

    model_pyo = problem_config.get_pyomo_model_core()

    # define the constraint
    # then immediately apply it to the model
    model_pyo.nchoosek = NChooseKConstraint(
        feature_keys=["x1", "x2", "x3", "x4", "x5"],
        min_count=min_count,
        max_count=max_count,
        none_also_valid=False,
    ).as_pyomo_constraint(model_pyo, problem_config.feat_list)

    # optimise the model
    params_pyomo = {"solver_name": "gurobi"}
    opt_pyo = PyomoOptimizer(problem_config, params=params_pyomo)
    res_pyo = opt_pyo.solve(enting, model_core=model_pyo)

    assert min_count <= sum(x > 1e-6 for x in res_pyo.opt_point) <= max_count


def test_constraint_list():
    problem_config = ProblemConfig(rnd_seed=73)
    build_reals_only_problem(problem_config)
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=50)
    testfunc_evals = eval_reals_only_testfunc(rnd_sample)

    enting = Enting(problem_config, params=PARAMS)
    # fit tree ensemble
    enting.fit(rnd_sample, testfunc_evals)

    model_pyo = problem_config.get_pyomo_model_core()

    # define the constraints
    constraints = [
        NChooseKConstraint(
            feature_keys=["x1", "x2", "x3", "x4", "x5"],
            min_count=1,
            max_count=3,
            none_also_valid=True,
        ),
        LinearInequalityConstraint(
            feature_keys=["x3", "x4", "x5"], coefficients=[1, 1, 1], rhs=10.0
        ),
    ]

    # apply constraints to the model
    model_pyo.problem_constraints = pyo.ConstraintList()
    ConstraintList(constraints).apply_pyomo_constraints(
        model_pyo, problem_config.feat_list, model_pyo.problem_constraints
    )

    # optimise the model
    params_pyomo = {"solver_name": "gurobi"}
    opt_pyo = PyomoOptimizer(problem_config, params=params_pyomo)
    res_pyo = opt_pyo.solve(enting, model_core=model_pyo)

    print(res_pyo.opt_point)
    assert 1 <= sum(x > 1e-6 for x in res_pyo.opt_point) <= 3
    assert sum(res_pyo.opt_point[2:]) < 10.0



import math
import random

import pytest

from entmoot import Enting, GurobiOptimizer, ProblemConfig
from entmoot.benchmarks import (
    build_multi_obj_categorical_problem,
    eval_multi_obj_cat_testfunc,
)
from entmoot.models.model_params import (
    EntingParams,
    TrainParams,
    TreeTrainParams,
    UncParams,
)


def get_leaf_center(bnds):
    return [
        (b[0] + b[1]) / 2 if not isinstance(b, set) else random.choice(list(b))
        for b in bnds
    ]


def run_gurobi(rnd_seed, n_obj, params, params_opt, num_samples=20, no_cat=False):
    # define benchmark problem
    problem_config = ProblemConfig(rnd_seed=rnd_seed)
    build_multi_obj_categorical_problem(problem_config, n_obj=n_obj, no_cat=no_cat)

    # sample data
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=num_samples)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=n_obj, no_cat=no_cat)

    # build model
    enting = Enting(problem_config, params=params)
    enting.fit(rnd_sample, testfunc_evals)

    # solve gurobi
    opt = GurobiOptimizer(problem_config, params=params_opt)

    x_sol, obj, mu, unc, leafs = opt.solve(enting)

    # predict enting model with solution
    x_sol_enc = problem_config.encode([x_sol])

    leaf_bnd = [
        enting.leaf_bnd_predict(f"obj_{obj}", opt.get_active_leaf_sol()[obj])
        for obj in range(n_obj)
    ]

    leaf_mid = [get_leaf_center(b) for b in leaf_bnd]

    mu_pred = [enting.mean_model.predict([m])[0][i] for i, m in enumerate(leaf_mid)]
    unc_pred = enting.unc_model.predict([x_sol_enc])[0]

    # compare model mean and uncertainty to prediction
    for m_opt, m_pred in zip(mu, mu_pred):
        assert math.isclose(
            m_opt, m_pred, abs_tol=1e-4
        ), f"`{m_opt}` and `{m_pred}` tree values are too small to test"

    assert (
        unc > 0.001 and unc_pred > 0.001
    ), f"`{unc}` and `{unc_pred}` are too small to test"
    assert math.isclose(
        unc, unc_pred, abs_tol=0.001
    ), f"`{unc}` and `{unc_pred}` unc values are not the same"


@pytest.mark.parametrize("dist_metric", ["l1", "l2", "euclidean_squared"])
@pytest.mark.parametrize("cat_metric", ["overlap", "of", "goodall4"])
@pytest.mark.parametrize("acq_sense", ["exploration"])
@pytest.mark.parametrize("rnd_seed", [100, 101, 102])
@pytest.mark.parametrize("n_obj", [1, 2])
def test_gurobi_consistency1(rnd_seed, n_obj, acq_sense, dist_metric, cat_metric):
    # define model params
    params = EntingParams(
        unc_params=UncParams(
            dist_metric=dist_metric,
            acq_sense=acq_sense,
            dist_trafo="normal",
            cat_metric=cat_metric,
        )
    )
    params_opt = {"LogToConsole": 1, "MIPGap": 0}
    run_gurobi(rnd_seed, n_obj, params, params_opt, num_samples=200)


@pytest.mark.parametrize("dist_metric", ["l1", "euclidean_squared"])
@pytest.mark.parametrize("cat_metric", ["overlap", "of", "goodall4"])
@pytest.mark.parametrize("acq_sense", ["penalty"])
@pytest.mark.parametrize("rnd_seed", [100, 101, 102])
@pytest.mark.parametrize("n_obj", [1, 2])
def test_gurobi_consistency2(rnd_seed, n_obj, acq_sense, dist_metric, cat_metric):
    # define model params
    params = EntingParams(
        unc_params=UncParams(
            dist_metric=dist_metric,
            acq_sense=acq_sense,
            dist_trafo="normal",
            cat_metric=cat_metric,
        )
    )
    params.unc_params.beta = 0.05
    params_opt = {"LogToConsole": 1, "MIPGap": 0}

    run_gurobi(rnd_seed, n_obj, params, params_opt, num_samples=300)


@pytest.mark.parametrize("dist_metric", ["l2"])
@pytest.mark.parametrize("cat_metric", ["overlap", "of", "goodall4"])
@pytest.mark.parametrize("acq_sense", ["penalty"])
@pytest.mark.parametrize("rnd_seed", [100, 101, 102])
@pytest.mark.parametrize("n_obj", [1, 2])
def test_gurobi_consistency3(rnd_seed, n_obj, acq_sense, dist_metric, cat_metric):
    # define model params
    params = EntingParams(
        unc_params=UncParams(
            dist_metric=dist_metric,
            acq_sense=acq_sense,
            dist_trafo="normal",
            cat_metric=cat_metric,
        ),
        # make tree model smaller to reduce testing time
        tree_train_params=TreeTrainParams(
            train_lib="lgbm",
            train_params=TrainParams(
                objective="regression",
                metric="rmse",
                boosting="gbdt",
                num_boost_round=2,
                max_depth=2,
                min_data_in_leaf=1,
                min_data_per_group=1,
                verbose=-1,
            ),
        ),
    )

    params.unc_params.beta = 0.05
    params_opt = {"LogToConsole": 1}

    if n_obj == 1:
        params_opt["MIPGap"] = 0
    else:
        # gurobi takes a long time to fully prove optimality here
        params_opt["MIPGapAbs"] = 0.001

    run_gurobi(rnd_seed, n_obj, params, params_opt, num_samples=300)


@pytest.mark.parametrize("dist_metric", ["l1", "l2", "euclidean_squared"])
@pytest.mark.parametrize("acq_sense", ["exploration"])
@pytest.mark.parametrize("rnd_seed", [100, 101, 102])
def test_gurobi_consistency4(rnd_seed, acq_sense, dist_metric):
    params = EntingParams(
        unc_params=UncParams(
            dist_metric=dist_metric,
            acq_sense=acq_sense,
            dist_trafo="standard",
        )
    )

    params.unc_params.beta = 0.1
    params_opt = {"LogToConsole": 1, "MIPGap": 1e-5}

    run_gurobi(rnd_seed, 1, params, params_opt, num_samples=20, no_cat=True)


@pytest.mark.parametrize("dist_metric", ["l1", "euclidean_squared"])
@pytest.mark.parametrize("acq_sense", ["penalty"])
@pytest.mark.parametrize("rnd_seed", [100, 101, 102])
def test_gurobi_consistency5(rnd_seed, acq_sense, dist_metric):
    params = EntingParams(
        unc_params=UncParams(
            dist_metric=dist_metric,
            acq_sense=acq_sense,
            dist_trafo="standard",
        )
    )
    params.unc_params.beta = 0.1
    params_opt = {"LogToConsole": 1, "MIPGap": 1e-5}

    run_gurobi(rnd_seed, 1, params, params_opt, num_samples=200, no_cat=True)


@pytest.mark.parametrize("dist_metric", ["l2"])
@pytest.mark.parametrize("acq_sense", ["penalty"])
@pytest.mark.parametrize("rnd_seed", [100, 101, 102])
def test_gurobi_consistency6(rnd_seed, acq_sense, dist_metric):
    params = EntingParams(
        unc_params=UncParams(
            dist_metric=dist_metric,
            acq_sense=acq_sense,
            dist_trafo="standard",
        )
    )
    params.unc_params.beta = 0.05
    params_opt = {"LogToConsole": 1, "MIPGapAbs": 0.01}

    run_gurobi(rnd_seed, 1, params, params_opt, num_samples=200, no_cat=True)



import pytest

from entmoot import ProblemConfig
from entmoot.benchmarks import (
    build_multi_obj_categorical_problem,
)


@pytest.mark.pipeline_test
def test_core_model_copy():
    def features_equal(model, model_copy, predicate):
        for f, f_copy in zip(model._all_feat, model_copy._all_feat):
            if type(f) is not type(f_copy):
                return False
            elif isinstance(f, dict):
                for v, v_copy in zip(f.values(), f_copy.values()):
                    if not predicate(v, v_copy):
                        return False
            else:
                if not predicate(f, f_copy):
                    return False
        return True

    # define problem
    problem_config = ProblemConfig(rnd_seed=73)
    # number of objectives
    build_multi_obj_categorical_problem(problem_config, n_obj=2)

    core_model_gurobi = problem_config.get_gurobi_model_core()
    core_model_gurobi_copy = problem_config.copy_gurobi_model_core(core_model_gurobi)

    assert len(core_model_gurobi.getVars()) == len(core_model_gurobi_copy.getVars())
    assert len(core_model_gurobi._all_feat) == len(core_model_gurobi_copy._all_feat)
    predicate = lambda u, v: all((u.lb == v.lb, u.ub == v.ub, u.v_type == v.v_type))
    assert features_equal(core_model_gurobi, core_model_gurobi_copy, predicate)

    core_model_pyomo = problem_config.get_pyomo_model_core()
    core_model_pyomo_copy = problem_config.copy_pyomo_model_core(core_model_pyomo)

    assert len(core_model_pyomo.x) == len(core_model_pyomo_copy.x)
    assert len(core_model_pyomo._all_feat) == len(core_model_pyomo_copy._all_feat)
    predicate = lambda u, v: all((u.lb == v.lb, u.ub == v.ub, u.domain == v.domain))
    assert features_equal(core_model_pyomo, core_model_pyomo_copy, predicate)



"""Test handling of objectives, and min/max problems"""

import numpy as np
import pytest
from pytest import approx

from entmoot import Enting, ProblemConfig, PyomoOptimizer
from entmoot.benchmarks import (
    build_multi_obj_categorical_problem,
    eval_multi_obj_cat_testfunc,
)


@pytest.mark.pipeline_test
def test_incorrect_shape_of_observations_raises_error():
    problem_config = ProblemConfig(rnd_seed=73)
    build_multi_obj_categorical_problem(problem_config, n_obj=1)

    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)

    params = {"unc_params": {"dist_metric": "l1", "acq_sense": "penalty"}}
    enting = Enting(problem_config, params=params)

    with pytest.raises(AssertionError):
        # too many objectives
        testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=2)
        enting.fit(rnd_sample, testfunc_evals)

    with pytest.raises(AssertionError):
        # too few observations
        testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=1)[:-3]
        enting.fit(rnd_sample, testfunc_evals)


@pytest.mark.pipeline_test
def test_max_predictions_equal_min_predictions():
    """The sign of the predicted objective is independent of max/min."""
    problem_config = ProblemConfig(rnd_seed=73)
    build_multi_obj_categorical_problem(problem_config, n_obj=1)
    problem_config.add_min_objective()

    problem_config_max = ProblemConfig(rnd_seed=73)
    build_multi_obj_categorical_problem(problem_config_max, n_obj=1)
    problem_config_max.add_max_objective()

    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=2)

    params = {"unc_params": {"dist_metric": "l1", "acq_sense": "exploration"}}
    enting = Enting(problem_config, params=params)
    enting.fit(rnd_sample, testfunc_evals)

    enting_max = Enting(problem_config_max, params=params)
    enting_max.fit(rnd_sample, testfunc_evals)

    sample = problem_config.get_rnd_sample_list(num_samples=3)
    pred = enting.predict(sample)
    pred_max = enting_max.predict(sample)

    for (m1, u1), (m2, u2) in zip(pred, pred_max):
        assert np.allclose(m1, m2, rtol=1e-5)
        assert np.allclose(u1, u2, rtol=1e-5)


@pytest.mark.pipeline_test
def test_max_objective_equals_minus_min_objective():
    """Assert that the solution found by the minimiser is the same as that of the maximiser for the negative objective function"""
    problem_config = ProblemConfig(rnd_seed=73)
    build_multi_obj_categorical_problem(problem_config, n_obj=1)

    problem_config_max = ProblemConfig(rnd_seed=73)
    build_multi_obj_categorical_problem(problem_config_max, n_obj=0)
    problem_config_max.add_max_objective()

    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=1)

    params = {"unc_params": {"dist_metric": "l1", "acq_sense": "penalty"}}
    enting = Enting(problem_config, params=params)
    enting.fit(rnd_sample, testfunc_evals)
    # pass negative test evaluations to the maximiser
    enting_max = Enting(problem_config, params=params)
    enting_max.fit(rnd_sample, -testfunc_evals)

    params_pyomo = {"solver_name": "gurobi"}
    res = PyomoOptimizer(problem_config, params=params_pyomo).solve(enting)
    res_max = PyomoOptimizer(problem_config_max, params=params_pyomo).solve(enting)

    assert res.opt_point == approx(res_max.opt_point, rel=1e-5)
    assert res.opt_val == approx(res_max.opt_val, rel=1e-3)

    # Note that OptResult.opt_val is *not* negated for maximization problems
    # but enting.predict(x) *is* negated.

    ((mu,), std) = enting.predict([res.opt_point])[0]
    ((mu_max,), std_max) = enting_max.predict([res.opt_point])[0]

    assert mu == approx(-mu_max, rel=1e-5)
    assert std == approx(std_max, rel=1e-5)



"""Integration tests on benchmark problems, comparing Pyomo and Gurobi"""

import random

import numpy as np
import pytest

from entmoot import Enting, GurobiOptimizer, ProblemConfig, PyomoOptimizer
from entmoot.benchmarks import (
    build_multi_obj_categorical_problem,
    eval_multi_obj_cat_testfunc,
)
from entmoot.models.model_params import EntingParams, UncParams


@pytest.mark.pipeline_test
def test_multiobj_constrained_problem():
    # define problem
    problem_config = ProblemConfig(rnd_seed=73)
    # number of objectives
    number_objectives = 2
    build_multi_obj_categorical_problem(problem_config, n_obj=number_objectives)

    # sample data
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=number_objectives)

    params = EntingParams(
        unc_params=UncParams(dist_metric="l1", acq_sense="exploration")
    )
    enting = Enting(problem_config, params=params)
    # fit tree ensemble
    enting.fit(rnd_sample, testfunc_evals)

    # Add constraints

    # Gurobi version
    # get optimization model
    model_gur = problem_config.get_gurobi_model_core()
    # extract decision variables
    x = model_gur._all_feat[3]
    y = model_gur._all_feat[4]
    z = model_gur._all_feat[5]
    # add constraint that all variables should coincide
    model_gur.addConstr(x == y)
    model_gur.addConstr(y == z)
    # Update model
    model_gur.update()

    # Build GurobiOptimizer object and solve optimization problem
    params_gurobi = {"MIPGap": 1e-3}
    opt_gur = GurobiOptimizer(problem_config, params=params_gurobi)

    res_gur = opt_gur.solve(enting, model_core=model_gur)
    x_opt = res_gur.opt_point[3]

    assert np.isclose(res_gur.opt_point[3:], x_opt, rtol=1.001).all()

    # Pyomo version
    import pyomo.environ as pyo

    model_pyo = problem_config.get_pyomo_model_core()
    # extract decision variables
    x = model_pyo._all_feat[3]
    y = model_pyo._all_feat[4]
    z = model_pyo._all_feat[5]
    # add constraint that all variables should coincide
    model_pyo.xy_equal_constr = pyo.Constraint(expr=x == y)
    model_pyo.yz_equal_constr = pyo.Constraint(expr=y == z)

    # Build GurobiOptimizer object and solve optimization problem
    params_pyomo = {"solver_name": "gurobi"}
    opt_pyo = PyomoOptimizer(problem_config, params=params_pyomo)

    res_pyo = opt_pyo.solve(enting, model_core=model_pyo)
    x_opt = res_pyo.opt_point[3]

    assert np.isclose(res_gur.opt_point[3:], x_opt, rtol=1.001).all()


@pytest.mark.pipeline_test
def test_simple_one_dimensional_problem():
    random.seed(42)

    def my_func(x: float) -> float:
        return x**2 + 1 + random.uniform(-0.2, 0.2)

    # Define a one-dimensional minimization problem with one real variable bounded by -2 and 3
    problem_config = ProblemConfig()
    problem_config.add_feature("real", (-2, 3))
    problem_config.add_min_objective()

    # Create training data using the randomly disturbed function f(x) = x^2 + 1 + eps
    X_train = np.reshape(np.linspace(-2, 3, 10), (-1, 1))
    y_train = np.reshape([my_func(x) for x in X_train], (-1, 1))

    # Define enting object and corresponding parameters
    params = EntingParams(
        unc_params=UncParams(dist_metric="l1", acq_sense="exploration")
    )
    enting = Enting(problem_config, params=params)
    # Fit tree model
    enting.fit(X_train, y_train)

    # Build PyomoOptimizer object with Gurobi as solver and solve optimization problem
    params_pyo = {"solver_name": "gurobi"}
    opt_pyo = PyomoOptimizer(problem_config, params=params_pyo)
    res = opt_pyo.solve(enting)

    # very little guarantee about position of next proposal
    # likely to be near center
    assert -1.5 < res.opt_point[0] < 1.5


@pytest.mark.pipeline_test
@pytest.mark.parametrize("dist_metric", ("l1", "l2", "euclidean_squared"))
@pytest.mark.parametrize("acq_sense", ("exploration", "penalty"))
@pytest.mark.parametrize("num_objectives", (1, 2))
def test_compare_pyomo_gurobipy_singleobj(dist_metric, acq_sense, num_objectives):
    """
    Ensures for a single objective example with l1 and l2 uncertainty metric and mixed feature types that optimization
    results for Gurobipy model and Pyomo model with Gurobi as optimizer coincide.
    """

    # define problem
    problem_config = ProblemConfig(rnd_seed=73)
    # number of objectives
    w = (0.4, 0.6) if num_objectives == 2 else None
    build_multi_obj_categorical_problem(problem_config, n_obj=num_objectives)

    # sample data
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=num_objectives)

    params = EntingParams(
        unc_params=UncParams(dist_metric=dist_metric, acq_sense=acq_sense)
    )
    enting = Enting(problem_config, params=params)
    # fit tree ensemble
    enting.fit(rnd_sample, testfunc_evals)

    # Build GurobiOptimizer object and solve optimization problem
    params_gurobi = {"NonConvex": 2, "MIPGap": 1e-3}
    opt_gur = GurobiOptimizer(problem_config, params=params_gurobi)
    res_gur = opt_gur.solve(enting, weights=w)

    # Build PyomoOptimizer object with Gurobi as solver and solve optimization problem
    params_pyo = {
        "solver_name": "gurobi",
        "solver_options": {"NonConvex": 2, "MIPGap": 1e-3},
    }
    opt_pyo = PyomoOptimizer(problem_config, params=params_pyo)
    res_pyo = opt_pyo.solve(enting, weights=w)

    # Compare optimal values (e.g. objective values) ...
    assert np.allclose(res_gur.opt_val, res_pyo.opt_val, rtol=1.001)
    # ... and optimal points (e.g. feature variables)
    assert np.allclose(res_gur.opt_point[2:], res_pyo.opt_point[2:], rtol=1.01)



import math

import numpy as np
import pytest

from entmoot import Enting, GurobiOptimizer, ProblemConfig
from entmoot.benchmarks import (
    build_multi_obj_categorical_problem,
    build_small_single_obj_categorical_problem,
    eval_multi_obj_cat_testfunc,
    eval_small_single_obj_cat_testfunc,
)
from entmoot.models.model_params import EntingParams, UncParams


def run_gurobi(
    rnd_seed,
    params,
    params_opt,
    num_samples,
    no_cat=False,
    rtol=1e-2,
    num_opt_samples=250000,
    smaller_problem=False,
):
    # define benchmark problem
    problem_config = ProblemConfig(rnd_seed=rnd_seed)
    if smaller_problem:
        build_small_single_obj_categorical_problem(problem_config, no_cat=no_cat)
    else:
        build_multi_obj_categorical_problem(problem_config, n_obj=1, no_cat=no_cat)

    # sample data
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=num_samples)
    if smaller_problem:
        testfunc_evals = eval_small_single_obj_cat_testfunc(rnd_sample, no_cat=no_cat)
    else:
        testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=1, no_cat=no_cat)

    # build model
    enting = Enting(problem_config, params=params)
    enting.fit(rnd_sample, testfunc_evals)

    # manual optimization estimate via sampling
    samples = problem_config.get_rnd_sample_numpy(num_samples=num_opt_samples)
    preds = enting.predict_acq(samples, is_enc=True)
    min_idx = np.argmin(preds)

    est_pred_sol = preds[min_idx][0]
    est_x_sol = samples[min_idx]

    # solve gurobi
    opt = GurobiOptimizer(problem_config, params=params_opt)

    x_sol, obj, mu, unc, leafs = opt.solve(enting)
    x_sol_enc = problem_config.encode([x_sol])

    # check that solutions are sufficiently close
    assert np.allclose(x_sol_enc, est_x_sol, rtol=rtol), (
        f"solutions deviate by more than {rtol} relative tolerance, "
        f"`{x_sol_enc}` vs. `{est_x_sol}`"
    )

    # check that optimal values are sufficiently close
    assert math.isclose(
        est_pred_sol, obj, abs_tol=1e-1
    ), f"`{est_pred_sol}` and `{obj}` optimal values are not the same"

    assert est_pred_sol >= obj, "estimated value is smaller than solver solution"

    # check that mu values are sufficiently close
    est_mu = enting.mean_model.predict([samples[min_idx]])[0]
    assert math.isclose(
        est_mu[0], mu[0], abs_tol=1e-1
    ), f"`{est_mu}` and `{mu}` mean values are not the same"

    # check that unc values are sufficiently large
    est_unc = enting.unc_model.predict([samples[min_idx]])[0]
    assert (
        unc > 0.001 and est_unc > 0.001
    ), f"`{unc}` and `{est_unc}` are too small to test"
    # check that mu values are sufficiently close
    est_unc = enting.unc_model.predict([samples[min_idx]])[0]
    assert math.isclose(
        est_unc, unc, abs_tol=1e-1
    ), f"`{est_unc}` and `{unc}` uncertainty values are not the same"


@pytest.mark.parametrize("dist_metric", ["l1", "l2", "euclidean_squared"])
@pytest.mark.parametrize("cat_metric", ["overlap", "of", "goodall4"])
@pytest.mark.parametrize("acq_sense", ["exploration"])
@pytest.mark.parametrize("rnd_seed", [100, 101])
def test_gurobi_optimality(rnd_seed, acq_sense, dist_metric, cat_metric):
    # define model params
    params = EntingParams(
        unc_params=UncParams(
            dist_metric=dist_metric,
            acq_sense=acq_sense,
            dist_trafo="normal",
            cat_metric=cat_metric,
        )
    )
    params_opt = {"LogToConsole": 1, "MIPGap": 0}
    run_gurobi(rnd_seed, params, params_opt, num_samples=20)



from typing import Sequence

import numpy as np
from numpy.typing import ArrayLike

from entmoot import ProblemConfig


def build_small_single_obj_categorical_problem(
    problem_config: ProblemConfig, no_cat=False
):
    """
    Builds a small test example which is frequently used by the tests.
    :param problem_config: ProblemConfig object where features and objectives are added
    :param n_obj: Number of objectives
    :return: None, the problem definition happens "inplace"
    """
    if not no_cat:
        problem_config.add_feature("categorical", ("blue", "orange", "gray"))
    problem_config.add_feature("integer", (5, 6))
    problem_config.add_feature("binary")
    problem_config.add_feature("real", (5.0, 6.0))

    problem_config.add_min_objective()


def eval_small_single_obj_cat_testfunc(X: ArrayLike, no_cat=False) -> np.ndarray:
    """
    Benchmark function with at least four input variables and one or two outputs
    :param X: Usually a numpy array or a list of tuples. Each row (or tuple) consists of at least four entries with the
    following properties:
    - The first one is a categorical variable with the possible values "blue", "orange" and "gray".
    - The second one is an integer variable.
    - The third one is a binary variable.
    - The fourth one and all additional variables are real numbers.
    :param n_obj: number of objectives (one or two)
    :return: objective value(s) for each data point
    """

    # without the dtype=object paramer, each entry of X is converted into a string
    X = np.array(X, dtype=object)

    def compute_objectives(xi: Sequence, no_cat=False):
        if no_cat:
            return (
                xi[1] * xi[2] * np.sin(sum(xi[3:]))
                + xi[1] * (1 - xi[2]) * np.cos(sum(xi[3:])),
                xi[1] * xi[2] * sum(xi[3:]) - xi[1] * (1 - xi[2]) * sum(xi[3:]),
            )

        if xi[0] == "blue":
            return (
                xi[1] * xi[2] * np.sin(sum(xi[3:]))
                + xi[1] * (1 - xi[2]) * np.cos(sum(xi[3:])),
                xi[1] * xi[2] * sum(xi[3:]) - xi[1] * (1 - xi[2]) * sum(xi[3:]),
            )
        elif xi[0] == "orange":
            return (
                xi[1] * xi[2] * (sum(xi[3:]) / len(xi[3:])) ** 2
                + xi[1] * (1 - xi[2]) * (sum(xi[3:]) / len(xi[3:])) ** 3,
                np.sqrt(
                    abs(xi[1] * xi[2] * sum(xi[3:]) + xi[1] * (1 - xi[2]) * sum(xi[3:]))
                ),
            )
        elif xi[0] == "gray":
            return (xi[2] * xi[3] ** xi[1], -(1 - xi[2]) * xi[3] ** xi[1])
        else:
            raise IOError(
                f"You provided the illegal value {xi[0]} for the categorical variable. Allowed values are "
                f"'blue', 'orange' and 'gray'"
            )

    return np.array([[sum(compute_objectives(xi, no_cat=no_cat)) / 2] for xi in X])


def build_multi_obj_categorical_problem(
    problem_config: ProblemConfig, n_obj: int = 2, no_cat=False
):
    """
    Builds a small test example which is frequently used by the tests.
    :param problem_config: ProblemConfig object where features and objectives are added
    :param n_obj: Number of objectives
    :return: None, the problem definition happens "inplace"
    """
    if not no_cat:
        problem_config.add_feature("categorical", ("blue", "orange", "gray"))
    problem_config.add_feature("integer", (5, 6))
    problem_config.add_feature("binary")
    problem_config.add_feature("real", (5.0, 6.0))
    problem_config.add_feature("real", (4.6, 6.0))
    problem_config.add_feature("real", (5.0, 6.0))

    for _ in range(n_obj):
        problem_config.add_min_objective()


def eval_multi_obj_cat_testfunc(
    X: ArrayLike, n_obj: int = 2, no_cat=False
) -> np.ndarray:
    """
    Benchmark function with at least four input variables and one or two outputs
    :param X: Usually a numpy array or a list of tuples. Each row (or tuple) consists of at least four entries with the
    following properties:
    - The first one is a categorical variable with the possible values "blue", "orange" and "gray".
    - The second one is an integer variable.
    - The third one is a binary variable.
    - The fourth one and all additional variables are real numbers.
    :param n_obj: number of objectives (one or two)
    :return: objective value(s) for each data point
    """

    # without the dtype=object paramer, each entry of X is converted into a string
    X = np.array(X, dtype=object)

    def compute_objectives(xi: Sequence, no_cat=False):
        if no_cat:
            return (
                xi[1] * xi[2] * np.sin(sum(xi[3:]))
                + xi[1] * (1 - xi[2]) * np.cos(sum(xi[3:])),
                xi[1] * xi[2] * sum(xi[3:]) - xi[1] * (1 - xi[2]) * sum(xi[3:]),
            )

        if xi[0] == "blue":
            return (
                xi[1] * xi[2] * np.sin(sum(xi[3:]))
                + xi[1] * (1 - xi[2]) * np.cos(sum(xi[3:])),
                xi[1] * xi[2] * sum(xi[3:]) - xi[1] * (1 - xi[2]) * sum(xi[3:]),
            )
        elif xi[0] == "orange":
            return (
                xi[1] * xi[2] * (sum(xi[3:]) / len(xi[3:])) ** 2
                + xi[1] * (1 - xi[2]) * (sum(xi[3:]) / len(xi[3:])) ** 3,
                np.sqrt(
                    abs(xi[1] * xi[2] * sum(xi[3:]) + xi[1] * (1 - xi[2]) * sum(xi[3:]))
                ),
            )
        elif xi[0] == "gray":
            return (xi[2] * xi[3] ** xi[1], -(1 - xi[2]) * xi[3] ** xi[1])
        else:
            raise IOError(
                f"You provided the illegal value {xi[0]} for the categorical variable. Allowed values are "
                f"'blue', 'orange' and 'gray'"
            )

    if n_obj == 2:
        return np.array([compute_objectives(xi, no_cat=no_cat) for xi in X])
    elif n_obj == 1:
        return np.array([[sum(compute_objectives(xi, no_cat=no_cat)) / 2] for xi in X])
    else:
        raise IOError(
            f"You provided the illegal value {n_obj} for the number of objectives. "
            f"Allowed values are 1 and 2"
        )


def build_reals_only_problem(problem_config: ProblemConfig):
    """A problem containing only real values, as used to demonstrate the NChooseK
    constraint.

    The minimum is (1.0, 2.0, 3.0, ...)"""

    problem_config.add_feature("real", (0.0, 5.0), name="x1")
    problem_config.add_feature("real", (0.0, 5.0), name="x2")
    problem_config.add_feature("real", (0.0, 5.0), name="x3")
    problem_config.add_feature("real", (0.0, 5.0), name="x4")
    problem_config.add_feature("real", (0.0, 5.0), name="x5")
    problem_config.add_min_objective()


def eval_reals_only_testfunc(X: ArrayLike):
    """The function (x1 - 1)**2 + (x2 - 2)**2 + ..."""
    x = np.atleast_2d(X)
    xbar = np.ones_like(x)
    xbar *= (np.arange(x.shape[1]) + 1)[None, :]
    y = np.sum((x - xbar) ** 2, axis=1)
    return y.reshape(-1, 1)
